%% FastFourierExtension_makefig4.m 
% This code generates "The Fast Slepian Transform" Figure 4. 
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

T = 1.5; % Extended interval half-length
cutoff = 1e-4; % Threshold for truncated pseudoinverse
alpha = 1e-8; % Regularization parameter for Tikhonov regularization
tolerance = 1e-5; % Tolerance for fast methods
M = 200; % Use 2M+1 Fourier series terms

% Function to reconstruct
% f(t) = sum a_k*exp((t-mu_k)/sigma_k) for random a_k, mu_k, sigma_k
rng(1);
K = 500;
mu = (2*rand(1,K)-1);
sigma = 0.1*rand(1,K)+0.01;
a = randn(1,K); a = a+0.5*sign(a);

% Sample f(t) at N equispaced points in [-1,1]
N = 2^15;
tsamp = (2*(0:N-1)'/N-1);
fsamp = 5*tsamp;
for k = 1:K,
    fsamp = fsamp + a(k)*exp(-abs((tsamp-mu(k))./sigma(k)));
end

% Sample f(t)*chi_[-1,1](t) at N equispaced points in [-T,T]
tsampT = T*(2*(0:N-1)'/N-1);
indexT = (abs(tsampT) <= 1);
fsampT = zeros(N,1);
fsampT(indexT) = 5*tsampT(indexT);
for k = 1:K,
    fsampT(indexT) = fsampT(indexT) + a(k)*exp(-abs((tsampT(indexT)-mu(k))./sigma(k)));
end

% Approximate f(t) using M Fourier series coefficients
fftf = fft(fsamp);
fhat = (sqrt(2)/N)*((-1).^(-M:M)').*[fftf((N-M+1):N);fftf(1:(M+1))];
fMsamp = (N/sqrt(2))*ifft([fhat((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);fhat(1:M).*((-1).^(-M:-1)')]);

% Compute Fourier Extension coefficients using slow pseudoinverse
fftfT = fft(fsampT);
fhatT = (sqrt(2*T)/N)*((-1).^(-M:M)').*[fftfT((N-M+1):N);fftfT(1:(M+1))];
Koverestimate = min(2*M+1,ceil((2*M+1)/T + log(2*(2*M+1))*log(2./cutoff)/pi^2+3));
[S,lambda] = myDPSS(2*M+1,1/(2*T),Koverestimate);
Kactual = nnz(lambda > cutoff);
ghatTpinv = S(:,1:Kactual)*((S(:,1:Kactual)'*fhatT)./lambda(1:Kactual));
gMsampTpinv = (N/sqrt(2*T))*ifft([ghatTpinv((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTpinv(1:M).*((-1).^(-M:-1)')]);
        
% Compute Fourier Extension coefficients using slow Tikhonov regularization
fftfT = fft(fsampT);
fhatT = (sqrt(2*T)/N)*((-1).^(-M:M)').*[fftfT((N-M+1):N);fftfT(1:(M+1))];
B = toeplitz(sinc((0:(2*M))/T)/T);
C = B^2;
C(1:(2*M+2):end) = C(1:(2*M+2):end) + alpha;
ghatTtik = C\(B*fhatT);
gMsampTtik = (N/sqrt(2*T))*ifft([ghatTtik((M+1):(2*M+1)).*((-1).^(0:M)');zeros(N-2*M-1,1);ghatTtik(1:M).*((-1).^(-M:-1)')]);

% Plot f(t) over [-1,1]
figure('Position', [100, 100, 800, 600]);
plot(tsamp,fsamp,'b-','linewidth',2)
ylim([floor(min(fsamp)),ceil(max(fsamp))])
xlim([-1,1])
set(gca,'XTick',-1:0.25:1);
set(gca,'FontSize',16)
title('Function $f(t)$','fontsize',22,'interpreter','latex'); 
xlabel('$t$','fontsize',22,'interpreter','latex');
ylabel('$f(t)$','fontsize',22,'interpreter','latex');

% Plot f(t) and f_M(t) over [-1,-1+delta] and [1-delta,1]
delta = 0.05;
ylimLeft = [-15,10];
ylimRight = [-15,10];
line_width = 2;
figure('Position', [100, 100, 800, 600]);
subplot(1,2,1)
plot(tsamp,fsamp,'b-',tsamp,fMsamp,'r-',tsampT(indexT),gMsampTpinv(indexT),'g-','linewidth',line_width)
ylim(ylimLeft)
xlim([-1,-1+delta])
h1 = legend('$f(t)$','$f_M(t)$','$g_M(t)$');
set(h1,'fontsize',16,'Location','east','interpreter','latex')
set(gca,'FontSize',16)
xlabel('$t$','fontsize',22,'interpreter','latex');
ylabel('$f(t)$','fontsize',22,'interpreter','latex');
subplot(1,2,2)
plot(tsamp,fsamp,'b-',tsamp,fMsamp,'r-',tsampT(indexT),gMsampTpinv(indexT),'g-','linewidth',line_width)
ylim(ylimRight)
xlim([1-delta,1])
h1 = legend('$f(t)$','$f_M(t)$','$g_M(t)$');
set(h1,'fontsize',16,'Location','west','interpreter','latex')
set(gca,'FontSize',16)
xlabel('$t$','fontsize',22,'interpreter','latex');
ylabel('$f(t)$','fontsize',22,'interpreter','latex');
set(gcf,'NextPlot','add');
axes;
h = title('Approximations of f(t) near endpoints','fontsize',22,'interpreter','latex');
set(gca,'Visible','off');
set(h,'Visible','on');