function FSP = FastSlepianProjection(N,W,cutoff,epsilon)
%% FastSlepianProjection.m 
% Function implementing the efficient projection onto the range of S_K as
% described in "The Fast Slepian Transform". The projection matrix onto the
% span of the first K \approx 2NW DPSSs can be approximated as the 
% prolate matrix plus a low rank correction matrix. 
% 
% Inputs:
% -N: Length of DPSSs
% -W: Half Bandwidth for DPSSs
% -cutoff: Projection onto the first K = cutoff DPSSs 
%          OR Projection onto the first K = #{ell : lambda_{N,W}^{(ell)} >= cutoff} DPSSs.
% -epsilon: Relative error in approximation
%
% Outputs:
% -FSP an object containing a Fast Slepian Projection function
%
% Given an Nx1 vector x, performing
% z = FSP.FSproj(x);
% results in z being an Nx1 vector such that ||z-S_KS_K^*x||_2 <= epsilon*||x||_2
% If x is an NxM array, then FSP.FSProj() operates columnwise.
%
% Notes:
% Using the FSproj() function is faster than using the FST() and FIST()
% functions in FastSlepianTransform.m and/or FastSlepianTransformMem.m, but
% this will not result in an intermediate compressed representation of x.
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

%% Fast Slepian Projection Function
FSP.FSproj = @FSproj;

%% Compute DPSSs and store transition region DPSSs
[FSP.S,lambda,lowerindex,upperindex] = transitionDPSS(N,W,epsilon,cutoff);
if(cutoff >= 1)
    K = cutoff;
    FSP.eig_weights = [ones(K-lowerindex+1,1);zeros(upperindex-K,1)]-lambda;
elseif(cutoff > eps)
    FSP.eig_weights = (lambda > cutoff)-lambda;
    K = nnz(lambda > cutoff)+lowerindex-1;
end

%% Precompuute length 2^q FFT of sinc function needed to apply B_{N,W} to a vector (where 2^q >= 2*N-1).
FSP.fftlength = 2^ceil(log2(2*N-1));
vecSinc = 2*W*sinc(2*W*(0:(N-1))');
FSP.sincFFT = fft([vecSinc;...
                   zeros(FSP.fftlength-2*N+1,1);...
                   vecSinc(end:-1:2)]);
               
%% Store signal length, bandwidth, tolerance, and number of DPSSs
FSP.N = N; % Signal Length
FSP.W = W; % Half-Bandwidth
FSP.eps = epsilon; % Tolerance
FSP.cutoff = cutoff;
FSP.K = K; % Number of DPSSs

%% Fast Slepian Projection
function Px = FSproj(x)
    if(FSP.N ~= size(x,1))
        disp('ERROR: Input has incorrect length')
    else
        % Toeplitz matrix multiply (Use FFT of length 2^q >= 2*N-1)
        Bx = ifft(bsxfun(@times,FSP.sincFFT,fft(x,FSP.fftlength,1)));
        
        % Low rank update
        SDSx = (FSP.S)*bsxfun(@times,FSP.eig_weights,(FSP.S)'*x);
        
        % Add Toeplitz*vector to low rank update
        Px = Bx(1:FSP.N,:)+SDSx;
    end
end

end