function FS = FastSlepianTransform(N,W,cutoff,epsilon)
%% FastSlepianTransform.m 
% Function implementing the efficient factorization of S_KS_K^* as
% described in "The Fast Slepian Transform". The projection matrix onto the
% span of the first K \approx 2NW DPSSs can be approximated as the 
% projection matrix onto the span of the \approx 2NW lowest frequency DFT 
% vectors plus a low rank correction matrix.
% 
% Inputs:
% -N: Length of DPSSs
% -W: Half Bandwidth for DPSSs
% -cutoff: Projection onto the first K = cutoff DPSSs 
%          OR Projection onto the first K = #{ell : lambda_{N,W}^{(ell)} >= cutoff} DPSSs.
% -epsilon: Relative error in approximation
%
% Outputs:
% -FS an object containing a Fast Slepian Transform function
%
% Given an Nx1 vector x, performing
% y = FS.FST(x);
% z = FS.FIST(y);
% results in y being a K'x1 compressed version of x where K' <= ceil(2NW)+[(12/pi^2)log(8N)+18]log(15/epsilon)
% and z is an Nx1 vector such that ||z-S_KS_K^*x||_2 <= epsilon*||x||_2
% If x is an NxM array, then FS.FST() and FS.FIST() operate columnwise.
%
% Notes: 
% - This code uses a modification of the math in "The Fast Slepian Transform"
% for which the matrices L_1 and L_2 are real instead of complex.
% - The values of delta_H, delta_A, and delta_B are chosen in a closer to
% optimal manner than they are in Section 2 of "The Fast Slepian Transform".
% - This version of the Fast Slepian Transform explicitly forms the
% matrices [L_1, U_1] and [L_2, U_2]. This version requires more memory and
% more precomputation time than FastSlepianTransformMem.m, but results in a
% faster projection time. 
%
% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

%% Fast Slepian Functions
FS.FST = @FST;
FS.FIST = @FIST;

%% Store signal length, bandwidth, tolerance, and number of DPSSs
FS.N = N; % Signal Length
FS.W = W; % Half-Bandwidth
FS.cutoff = cutoff; % Number of DPSSs
FS.epsilon = epsilon; % Tolerance

%% Error on individual terms (somewhat close to optimal)
epsilonF = 2*epsilon/3; % Error in approximating B_{N,W}-F_{N,W}F_{N,W}^*
mu = 2*log(8*N-4)/pi^2+1/log(2)+1/log(6/pi);
deltah = (log(8*N-4)/pi)*epsilonF/mu;
deltaa = (1/log(2))*epsilonF/mu;
deltab = (1/log(6/pi))*epsilonF/mu;
epsilonS = epsilon/3; % Error in approximating B_{N,W}-S_KS_K^*             

%% Compute DPSSs and store transition region DPSSs
[S,lambda,lowerindex,upperindex] = transitionDPSS(N,W,epsilonS,cutoff);
if(cutoff >= 1)
    K = cutoff;
    eig_weights = [ones(K-lowerindex+1,1);zeros(upperindex-K,1)]-lambda;
elseif(cutoff > eps)
    eig_weights = (lambda > cutoff)-lambda;
end

%% Diagonal matrices (store as column vectors)
R = round(W*N-1/2-1e-3);
W1 = (R+1/2)/N;
vecn = ((-(N-1)/2):((N-1)/2))';
DAcos = cos((2*pi*W1)*vecn);
DAsin = sin((2*pi*W1)*vecn);
DBcos = cos((pi*(W+W1))*vecn);
DBsin = sin((pi*(W+W1))*vecn);

%% Approximate Scaled Hilbert Matrix (1/pi)*H ~~ Z*Z'
rh = ceil(log(8*N-4)*log(4*pi/deltah)/pi^2);

k1 = 1/(2*N-1);
k = sqrt(1-k1^2);
Kk = ellipticK(k^2);
[~,~,p] = ellipj((1:2:(2*rh-1))*Kk/(2*rh),k^2);
p = (N-1/2)*p;

Z = zeros(N,rh);
vecn = ((1/2):(N-1/2))';
Z(:,1) = sqrt(2*p(1)/pi)*ones(N,1)./(vecn+p(1));
for j = 2:rh,
    Z(:,j) = sqrt(p(j)/p(j-1))*(1-(p(j)+p(j-1))./(vecn+p(j))).*(Z(:,j-1));
end

%% Approximate A1 ~~ Va*Ca*Va' and B0 ~~ Vb*Cb*Vb'
ra = ceil(log(2/(3*pi*deltaa))/(2*log(2)));
while((2/(3*pi))*sqrt(2/((4*ra-1)*4*ra))/2^(2*ra-2) <= deltaa)
    ra = ra-1;
end

vecZeta = zeta(2:2:(2*ra));
vecCoeff = (2/(N*pi))*(1-(1-2.^(-(1:2:(2*ra-1)))).*vecZeta);
vecGamma = gamma(1:(2*ra));
Ca = zeros(2*ra,2*ra);
for k = 1:ra,
    for l = 0:(2*k-1),
        Ca(l+1,2*k-l) = (-1)^(2*k-1-l)*vecCoeff(k).*vecGamma(2*k)./(vecGamma(l+1)*vecGamma(2*k-l));
    end
end

rb = 0;
while((2/pi)*sqrt(2/((4*rb+1)*(4*rb+2)))*(pi*abs(W-W1)*N)^(2*rb+1)/gamma(2*rb+2) > deltab)
    rb = rb+1;
end

vecGamma = gamma(1:(2*rb));
vecCoeff = (2/(N*pi))*(-1).^(0:(rb-1)).*(pi*(W-W1)*N).^(1:2:(2*rb-1))./gamma(2:2:(2*rb));
Cb = zeros(2*rb-1,2*rb-1);
for k = 0:(rb-1),
    for l = 0:(2*k),
        Cb(l+1,2*k-l+1) = (-1)^(2*k-l)*vecCoeff(k+1).*vecGamma(2*k+1)./(vecGamma(l+1)*vecGamma(2*k+1-l));
    end
end

% Va and Vb share many columns
V = ones(N,max(2*ra,2*rb-1));
vecn = (0:(1/N):(1-1/N))';
for k = 1:max(2*ra-1,2*rb-2),
    V(:,k+1) = V(:,k).*vecn;
end

%% Form FST matrices
DAsinZ = bsxfun(@times,DAsin,Z);
DAcosZ = bsxfun(@times,DAcos,Z);
DAsinVa = bsxfun(@times,DAsin,V(:,1:(2*ra)));
DAcosVa = bsxfun(@times,DAcos,V(:,1:(2*ra)));
DBsinVb = bsxfun(@times,DBsin,V(:,1:(2*rb-1)));
DBcosVb = bsxfun(@times,DBcos,V(:,1:(2*rb-1)));
FS.L1U1 = [DAsinZ,DAcosZ,DAsinZ(end:-1:1,:),DAcosZ(end:-1:1,:),DAsinVa*Ca,-DAcosVa*Ca,DBcosVb*Cb,DBsinVb*Cb,bsxfun(@times,S,eig_weights')];
FS.L2U2 = [DAcosZ(end:-1:1,:),DAsinZ(end:-1:1,:),DAcosZ,DAsinZ,DAcosVa,DAsinVa,DBcosVb,DBsinVb,S]';

%% Store a few more parameters
FS.R = R; % Use 2R+1 DFT vectors
FS.K1 = size(FS.L1U1,2)+2*R+1; % Total number of coefficients needed.

%% Fast Slepian Transform
function y = FST(x)
    if(FS.N ~= size(x,1))
        disp('ERROR: Input has incorrect length')
    else
        N = FS.N;
        R = FS.R;
        Fx = fft(x,N,1);
        y = [Fx([1:(R+1),(N-R+1):N],:);(FS.L2U2)*x];
    end
end

%% Fast Inverse Slepian Transform
function z = FIST(y)
    [K1,M] = size(y);
    if(FS.K1 ~= K1)
        disp('ERROR: Input has incorrect length')
    else
        N = FS.N;
        R = FS.R;
        z = ifft([y(1:(R+1),:);zeros(N-2*R-1,M);y((R+2):(2*R+1),:)],N,1)...
            +(FS.L1U1)*y((2*R+2):end,:);        
    end
end

end
