%% Test_FST_FSP.m 
% Code to compare the speed of the Fast Slepian Transform/Projection to the
% exact Slepian projection. This code generates the data for "The Fast
% Slepian Transform" Figures 2 and 3. 
% After running this script, run Test_FST_FSP_makefig.m to generate the top 
% panels of Figures 2 and 3. You will need to rerun this code with W = 1/16 
% and W = 1/64 to generate the middle and bottom panels of Figures 2 and 3.

% Most recent change - 08/07/2017
%
% Copyright 2017, Santhosh Karnik
%
% This file is part of Fast Slepian Transform (FST) Toolbox version 1.0.
%
%    FST is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    FST is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with FST.  If not, see <http://www.gnu.org/licenses/>.
%
% If you use this code in an academic paper, please cite our paper:
% S. Karnik, Z. Zhu, M. B. Wakin, J. K. Romberg, and M. A. Davenport., 
% The fast Slepian transform, Appl. Comput. Harmon. Anal. (2017), 
% http://dx.doi.org/10.1016/j.acha.2017.07.005

% Values of N (signal length) to test
vecN = sort([4*2.^(5:17),5*2.^(5:17),6*2.^(5:17),7*2.^(5:17),2^20]);
W = 1/4; % Fixed value of W (half-bandwidth)
Wstring = '2500'; % Stuff for filename
vecK = round(2*vecN*W); % Values of K (number of DPSSs to project onto)
vecEpsilon = 10.^(-(3:3:12)'); % Values of epsilon to test
Nmax = 3*2^11/sqrt(W); % Don't test exact projection when N > Nmax. 
% For large values of N, calling [S,lambda] = dpss(N,N*W,K); may yield an
% out of memory error and/or may cause your computer to freeze. The best
% value of Nmax is machine dependent. Choose this wisely. 

num_trials = 25; % Repeat tests num_trials times to get a good average time
num_vectors = 40; % Number of vectors to project each trial

% Create arrays for timing statistics
Exact_setup_time = zeros(1,length(vecN(vecN <= Nmax)),num_trials);
Exact_fwd_time = zeros(1,length(vecN(vecN <= Nmax)),num_trials);
Exact_inv_time = zeros(1,length(vecN(vecN <= Nmax)),num_trials);
FST_setup_time = zeros(length(vecEpsilon),length(vecN),num_trials);
FST_fwd_time = zeros(length(vecEpsilon),length(vecN),num_trials);
FST_inv_time = zeros(length(vecEpsilon),length(vecN),num_trials);
FST_rank = zeros(length(vecEpsilon),length(vecN));
FSP_setup_time = zeros(length(vecEpsilon),length(vecN),num_trials);
FSP_proj_time = zeros(length(vecEpsilon),length(vecN),num_trials);
FSP_rank = zeros(length(vecEpsilon),length(vecN));

% For each trial
for trial = 1:num_trials,
    % For each value of N
    for jj = [1,2,randperm(length(vecN))],
        N = vecN(jj);
        K = vecK(jj);
        x = randn(N,num_vectors)+1i*randn(N,num_vectors);
        
        disp([trial, N])
        
        % Test Naive Projection iff N <= Nmax
        if(N <= Nmax)
            tic;
            Sk = dpss(N,N*W,K);
            Exact_setup_time(1,jj,trial) = toc;
            
            y = zeros(K,num_vectors);
            tic;
            for m = 1:num_vectors,
                y(:,m) = Sk'*x(:,m);
            end
            Exact_fwd_time(1,jj,trial) = toc/num_vectors;
            
            z = zeros(N,num_vectors);
            tic;
            for m = 1:num_vectors,
                z(:,m) = Sk*y(:,m);
            end
            Exact_inv_time(1,jj,trial) = toc/num_vectors;
            
            clear Sk y z
        end
        
        % For each value of epsilon
        for ii = randperm(4),
            epsilon = vecEpsilon(ii);
            
            % Test Fast Slepian Transform
            tic;
            FS = FastSlepianTransform(N,W,K,epsilon);
            FST_setup_time(ii,jj,trial) = toc;

            y = zeros(FS.K1,num_vectors);
            tic;
            for m = 1:num_vectors,
                y(:,m) = FS.FST(x(:,m));
            end
            FST_fwd_time(ii,jj,trial) = toc/num_vectors;
            
            z = zeros(N,num_vectors);
            tic;
            for m = 1:num_vectors,
                z(:,m) = FS.FIST(y(:,m));
            end
            FST_inv_time(ii,jj,trial) = toc/num_vectors;

            if(trial == 1)
                FST_rank(ii,jj) = FS.K1-(2*FS.R+1);
            end
            clear FS2 y z
            
            % Test Fast Slepian Projection
            tic;
            FSP = FastSlepianProjection(N,W,K,epsilon);
            FSP_setup_time(ii,jj,trial) = toc;

            z = zeros(N,num_vectors);
            tic;
            for m = 1:num_vectors,
                z(:,m) = FSP.FSproj(x(:,m));
            end
            FSP_proj_time(ii,jj,trial) = toc/num_vectors;
            
            if(trial == 1)
                FSP_rank(ii,jj) = length(FSP.eig_weights);
            end
            clear FSP z
        end
    end
    
    clear x
% Save intermediate results (optional)
%     save('Test_incomplete.mat')
%     pause(15)
end

clear x trial ii jj N K epsilon

% Compute averages over all trials
Exact_setup_avg_time = mean(Exact_setup_time,3);
Exact_fwd_avg_time = mean(Exact_fwd_time,3);
Exact_inv_avg_time = mean(Exact_inv_time,3);
FST_setup_avg_time = mean(FST_setup_time,3);
FST_fwd_avg_time = mean(FST_fwd_time,3);
FST_inv_avg_time = mean(FST_inv_time,3);
FSP_setup_avg_time = mean(FSP_setup_time,3);
FSP_proj_avg_time = mean(FSP_proj_time,3);
Exact_proj_avg_time = Exact_fwd_avg_time + Exact_inv_avg_time;
FST_proj_avg_time = FST_fwd_avg_time + FST_inv_avg_time;

% Save data
save(['Test_FST_FSP_W',Wstring,'.mat'])